@@ -8,11 +8,11 @@ from datetime import datetime, timedelta
 import json
 import platform
 import yaml
-import six
 import gzip
 import functools
 import time
 import requests
+from werkzeug.exceptions import BadRequest
 
 # HACK: restore first two entries of sys path after wandb load
 save_path = sys.path[:2]
@@ -20,7 +20,7 @@ import wandb
 
 sys.path[0:0] = save_path
 import logging
-from six.moves import urllib
+import urllib
 import threading
 
 RequestsMock = None
@@ -61,6 +61,7 @@ def default_ctx():
         "artifacts": {},
         "artifacts_by_id": {},
         "artifacts_created": {},
+        "portfolio_links": {},
         "upsert_bucket_count": 0,
         "out_of_date": False,
         "empty_query": False,
@@ -175,7 +176,11 @@ def run(ctx):
         "events": ['{"cpu": 10}', '{"cpu": 20}', '{"cpu": 30}'],
         "files": {
             # Special weights url by default, if requesting upload we set the name
-            "edges": [{"node": fileNode,}]
+            "edges": [
+                {
+                    "node": fileNode,
+                }
+            ]
         },
         "sampledHistory": [[{"loss": 0, "acc": 100}, {"loss": 1, "acc": 0}]],
         "shouldStop": False,
@@ -230,12 +235,14 @@ def artifact(
                 "alias": "v%i" % ctx["page_count"],
             }
         ],
-        "artifactSequence": {"name": collection_name,},
+        "artifactSequence": {
+            "name": collection_name,
+        },
         "artifactType": {"name": "dataset"},
         "currentManifest": {
             "file": {
                 "directUrl": request_url_root
-                + "/storage?file=wandb_manifest.json&id={}".format(_id)
+                + f"/storage?file=wandb_manifest.json&id={_id}"
             }
         },
     }
@@ -254,7 +261,7 @@ def paginated(node, ctx, extra={}):
     }
 
 
-class CTX(object):
+class CTX:
     """This is a silly threadsafe wrapper for getting ctx into the server
     NOTE: This will stop working for live_mock_server if we make pytest run
     in parallel.
@@ -411,7 +418,7 @@ class SnoopRelay:
                 ctx["run_ids"].append(run_id)
             run_ctx = ctx["runs"].setdefault(run_id, default_ctx())
 
-            # NOTE: not used, added for consistancy with non-relay mode
+            # NOTE: not used, added for consistency with non-relay mode
             r = run_ctx.setdefault("run", {})
             r.setdefault("display_name", f"relay_name-{run_num}")
             r.setdefault("storage_id", f"storageid{run_num}")
@@ -420,7 +427,7 @@ class SnoopRelay:
 
             # TODO: handle errors better
             try:
-                # NOTE: We are using wandb but it isnt a strict depenedency
+                # NOTE: We are using wandb, but it isn't a strict dependency
                 import wandb
 
                 api = wandb.Api()
@@ -475,7 +482,13 @@ def create_app(user_ctx=None):
     def update_ctx():
         """Updating context for live_mock_server"""
         ctx = get_ctx()
-        body = request.get_json()
+        # in Flask/Werkzeug 2.1.0 get_json raises an exception on
+        # empty json, so we try/catch here
+        try:
+            body = request.get_json()
+        except BadRequest:
+            body = None
+
         if request.method == "GET":
             ctx = snoop.context_enrich(ctx)
             return json.dumps(ctx)
@@ -755,7 +768,11 @@ def create_app(user_ctx=None):
                 return json.dumps(
                     {
                         "data": {
-                            "QueryType": {"fields": [{"name": "serverInfo"},]},
+                            "QueryType": {
+                                "fields": [
+                                    {"name": "serverInfo"},
+                                ]
+                            },
                             "ServerInfoType": {
                                 "fields": [
                                     {"name": "cliVersionInfo"},
@@ -769,7 +786,11 @@ def create_app(user_ctx=None):
             return json.dumps(
                 {
                     "data": {
-                        "QueryType": {"fields": [{"name": "serverInfo"},]},
+                        "QueryType": {
+                            "fields": [
+                                {"name": "serverInfo"},
+                            ]
+                        },
                         "ServerInfoType": {
                             "fields": [
                                 {"name": "cliVersionInfo"},
@@ -858,7 +879,15 @@ def create_app(user_ctx=None):
             )
         if "mutation CreateAgent(" in body["query"]:
             return json.dumps(
-                {"data": {"createAgent": {"agent": {"id": "mock-server-agent-93xy",}}}}
+                {
+                    "data": {
+                        "createAgent": {
+                            "agent": {
+                                "id": "mock-server-agent-93xy",
+                            }
+                        }
+                    }
+                }
             )
         if "mutation Heartbeat(" in body["query"]:
             new_run_needed = body["variables"]["runState"] == "{}"
@@ -868,7 +897,9 @@ def create_app(user_ctx=None):
                 {
                     "data": {
                         "agentHeartbeat": {
-                            "agent": {"id": "mock-server-agent-93xy",},
+                            "agent": {
+                                "id": "mock-server-agent-93xy",
+                            },
                             "commands": (
                                 json.dumps(
                                     [
@@ -900,8 +931,8 @@ def create_app(user_ctx=None):
             run_ctx = ctx["runs"].setdefault(run_id, default_ctx())
 
             r = run_ctx.setdefault("run", {})
-            r.setdefault("display_name", "lovely-dawn-{}".format(run_num + 32))
-            r.setdefault("storage_id", "storageid{}".format(run_num))
+            r.setdefault("display_name", f"lovely-dawn-{run_num + 32}")
+            r.setdefault("storage_id", f"storageid{run_num}")
             r.setdefault("project_name", "test")
             r.setdefault("entity_name", "mock_server_entity")
 
@@ -1006,12 +1037,15 @@ def create_app(user_ctx=None):
                     }
                 )
             return json.dumps({"data": {"prepareFiles": {"files": {"edges": nodes}}}})
+        if "mutation LinkArtifact(" in body["query"]:
+            if ART_EMU:
+                return ART_EMU.link(variables=body["variables"])
         if "mutation CreateArtifact(" in body["query"]:
             if ART_EMU:
                 return ART_EMU.create(variables=body["variables"])
 
             collection_name = body["variables"]["artifactCollectionNames"][0]
-            app.logger.info("Creating artifact {}".format(collection_name))
+            app.logger.info(f"Creating artifact {collection_name}")
             ctx["artifacts"] = ctx.get("artifacts", {})
             ctx["artifacts"][collection_name] = ctx["artifacts"].get(
                 collection_name, []
@@ -1040,7 +1074,14 @@ def create_app(user_ctx=None):
             art = artifact(ctx, id_override=id)
             if len(art.get("aliases", [])) and not delete_aliases:
                 raise Exception("delete_aliases not set, but artifact has aliases")
-            return {"data": {"deleteArtifact": {"artifact": art, "success": True,}}}
+            return {
+                "data": {
+                    "deleteArtifact": {
+                        "artifact": art,
+                        "success": True,
+                    }
+                }
+            }
         if "mutation CreateArtifactManifest(" in body["query"]:
             manifest = {
                 "id": 1,
@@ -1061,7 +1102,13 @@ def create_app(user_ctx=None):
             run_ctx = ctx["runs"].setdefault(run_name, default_ctx())
             for c in ctx, run_ctx:
                 c["manifests_created"].append(manifest)
-            return {"data": {"createArtifactManifest": {"artifactManifest": manifest,}}}
+            return {
+                "data": {
+                    "createArtifactManifest": {
+                        "artifactManifest": manifest,
+                    }
+                }
+            }
         if "mutation UpdateArtifactManifest(" in body["query"]:
             manifest = {
                 "id": 1,
@@ -1078,7 +1125,13 @@ def create_app(user_ctx=None):
                     "uploadHeaders": "",
                 },
             }
-            return {"data": {"updateArtifactManifest": {"artifactManifest": manifest,}}}
+            return {
+                "data": {
+                    "updateArtifactManifest": {
+                        "artifactManifest": manifest,
+                    }
+                }
+            }
         if "mutation CreateArtifactFiles" in body["query"]:
             if ART_EMU:
                 return ART_EMU.create_files(variables=body["variables"])
@@ -1235,7 +1288,9 @@ def create_app(user_ctx=None):
                     if full_name is not None:
                         collection_name = full_name.split(":")[0]
                     art = artifact(
-                        ctx, collection_name=collection_name, request_url_root=base_url,
+                        ctx,
+                        collection_name=collection_name,
+                        request_url_root=base_url,
                     )
                 # code artifacts use source-RUNID names, we return the code type
                 art["artifactType"] = {"id": 2, "name": "code"}
@@ -1781,10 +1836,14 @@ def create_app(user_ctx=None):
                     },
                 }
         elif file == "wandb-metadata.json":
+            if ctx["return_jupyter_in_run_info"]:
+                code_path = "one_cell.ipynb"
+            else:
+                code_path = "train.py"
             result = {
                 "docker": "test/docker",
                 "program": "train.py",
-                "codePath": "train.py",
+                "codePath": code_path,
                 "args": ["--test", "foo"],
                 "git": ctx.get("git", {}),
             }
@@ -1796,7 +1855,7 @@ def create_app(user_ctx=None):
         elif file == "diff.patch":
             # TODO: make sure the patch is valid for windows as well,
             # and un skip the test in test_cli.py
-            return """
+            return r"""
 diff --git a/patch.txt b/patch.txt
 index 30d74d2..9a2c773 100644
 --- a/patch.txt
@@ -1807,7 +1866,7 @@ index 30d74d2..9a2c773 100644
 +testing
 \ No newline at end of file
 """
-        # emulate azure when we're receving requests from them
+        # emulate azure when we're receiving requests from them
         if extra is not None:
             return "", 201
         return "", 200
@@ -1983,13 +2042,13 @@ index 30d74d2..9a2c773 100644
 
     @app.errorhandler(404)
     def page_not_found(e):
-        print("Got request to: %s (%s)" % (request.url, request.method))
+        print(f"Got request to: {request.url} ({request.method})")
         return "Not Found", 404
 
     return app
 
 
-RE_DATETIME = re.compile("^(?P<date>\d+-\d+-\d+T\d+:\d+:\d+[.]\d+\s)(?P<rest>.*)$")
+RE_DATETIME = re.compile(r"^(?P<date>\d+-\d+-\d+T\d+:\d+:\d+[.]\d+\s)(?P<rest>.*)$")
 
 
 def strip_datetime(s):
@@ -2000,7 +2059,7 @@ def strip_datetime(s):
     return s
 
 
-class ParseCTX(object):
+class ParseCTX:
     def __init__(self, ctx, run_id=None):
         self._ctx = ctx["runs"][run_id] if run_id else ctx
         self._run_id = run_id
@@ -2012,21 +2071,21 @@ class ParseCTX(object):
             files = update.get("files")
             if not files:
                 continue
-            for k, v in six.iteritems(files):
+            for k, v in files.items():
                 data.setdefault(k, []).append(v)
         return data
 
     def get_filestream_file_items(self):
         data = {}
         fs_file_updates = self.get_filestream_file_updates()
-        for k, v in six.iteritems(fs_file_updates):
+        for k, v in fs_file_updates.items():
             l = []
             for d in v:
                 offset = d.get("offset")
                 content = d.get("content")
                 assert offset is not None
                 assert content is not None
-                # this check isnt valid right now.
+                # this check isn't valid right now.
                 # TODO: lets just assume it is fine, look into this later
                 # assert offset == 0 or offset == len(l), (k, v, l, d)
                 if not offset:
@@ -2162,6 +2221,10 @@ class ParseCTX(object):
     def artifacts(self):
         return self._ctx.get("artifacts_created") or {}
 
+    @property
+    def portfolio_links(self):
+        return self._ctx.get("portfolio_links") or {}
+
     @property
     def tags(self):
         return self._ctx.get("tags") or []
